/*
 * *****************************************************************
 * *                                                               *
 * *    Copyright (c) Digital Equipment Corporation, 1991, 1996    *
 * *                                                               *
 * *   All Rights Reserved.  Unpublished rights  reserved  under   *
 * *   the copyright laws of the United States.                    *
 * *                                                               *
 * *   The software contained on this media  is  proprietary  to   *
 * *   and  embodies  the  confidential  technology  of  Digital   *
 * *   Equipment Corporation.  Possession, use,  duplication  or   *
 * *   dissemination of the software and media is authorized only  *
 * *   pursuant to a valid written license from Digital Equipment  *
 * *   Corporation.                                                *
 * *                                                               *
 * *   RESTRICTED RIGHTS LEGEND   Use, duplication, or disclosure  *
 * *   by the U.S. Government is subject to restrictions  as  set  *
 * *   forth in Subparagraph (c)(1)(ii)  of  DFARS  252.227-7013,  *
 * *   or  in  FAR 52.227-19, as applicable.                       *
 * *                                                               *
 * *****************************************************************
 */
/*
 * HISTORY
 */
/*
 */
/*	
 * @(#)$RCSfile: atty.h,v $ $Revision: 1.2.6.3 $ (DEC I18N) $Date: 1995/11/09 20:41:00 $
 */ 
/*
 * atty.h
 *
 * Revision History:
 *
 * 10-Mar-92	W.M. Long
 *      First draft - this header file specifies all constants and data
 *	data structure used by the Asian BSD line discipline.
 */

#ifndef	_SYS_ATTY_H_
#define _SYS_ATTY_H_

#include <sys/tty.h>
#include <sys/uio.h>
#include <sys/clist.h>
#include <sys/aioctl.h>

/*====================[ LITERALS & MACROS DEFINITIONS ]======================*/

/*
 * Macros to set, clear and test flag bits
 */
#define	SET_FLAG(flag, bits)	((flag) |=  (bits))
#define	CLR_FLAG(flag, bits)	((flag) &= ~(bits))
#define	TST_FLAG(flag, bits)	((flag) &   (bits))

/*
 * C1 character definitions for ODL operation
 */
#define	CHAR_DCS	0x90
#define	CHAR_ST		0x9c

/*
 * Wide character list constants
 * WCBSIZE = number of wide characters in each wcblock
 */
#define TTY_WCHARMASK	0xffffff	  /* Wide character mask	*/
#define WCINFO_OFFSET	sizeof(Wchar_t *) /* sizeof(wc_next)		*/
#define	WCBSIZE		((CBLOCK - sizeof(Wchar_t *))/sizeof(Wchar_t))
#define WCROUND		CROUND

/*
 * Flags assignment in the status byte (MSB) of the wide character
 */
#define	TTY_WQUOTE	0x08000000	/* Character has been quoted	*/
#define	WDELAY_FLAG	0x10000000	/* Indicate a delay character	*/

/*
 * A cbuffer structure is used to hold input character to be fed through the
 * mbtowc routine. The c_ptr pointer points to the beginning of valid data.
 */
#define	CBUFSIZ		16	/* Size of character input buffer */

/*
 * Constants for history structure
 */
#define	HISTMAX		32	/* Number of history entries	*/
#define	HISTLEN		128	/* Max # of chars per entry	*/
#define	HISTBUF		(HISTMAX * HISTLEN / 4)
#define	HISTBEG		1	/* Beginning of history buffer	*/
#define	HISTEND		HISTBUF	/* End of history buffer	*/
#define	AHS_LITERAL	0x01

/*
 * Constants for selecting terminal and application code
 */
#define	USE_TERMCODE	0	/* Use terminal code	*/
#define	USE_APPLCODE	1	/* Use application code	*/

/*
 * Aliases for queues and buffers
 */
#define RAWQ		(tp)->t_rawq
#define CANQ		(tp)->t_canq
#define OUTQ		(tp)->t_outq
#define WRAWQ		(atp)->at_wrawq
#define WCANQ		(atp)->at_wcanq
#define WOUTQ		(atp)->at_woutq
#define	WPENDQ		(atp)->at_wpendq
#define	WPREHQ		(atp)->at_wprehq
#define	WDISPQ		(atp)->at_wdispq
#define	WERAWQ		(atp)->at_werawq
#define	WECANQ		(atp)->at_wecanq
#define	ASTIB		(atp)->at_astib
#define	AINPB		(atp)->at_ainpb
#define AOUTB		(atp)->at_aoutb
#define	TRAWB		(atp)->at_trawb
#define	TOUTB		(atp)->at_toutb

/*
 * Generic atty state
 */
#define	AT_WAITING	0x0100	/* Waiting in ttywait() */
#define	AT_ISOPEN	0x0200
#define	AT_HISTORY	0x0400
#define	AT_LITERAL	0x0800
#define	AT_RUB_CLEAR	0x1000

/*
 * Code conversion states
 */
#define	CC_HIRAGANA_MARU 0x0001	/* Previous char is Hiragana maru	*/
#define	CC_HIRAGANA_TEN	 0x0002	/* Previous char is Hiragana ten	*/
#define	CC_INKANJI	 0x0004	/* Input Kanji (for jis7 & jis8)	*/
#define	CC_INKANA	 0x0008	/* Input 1-byte kana (for jis7 & jis8)	*/
#define	CC_OUTKANJI	 0x0010	/* Output Kanji (for jis7 & jis8)	*/
#define	CC_CHAR_PENDING	 0x0020	/* Character pending in cc_lastwc	*/
#define	CC_ADECKANJI	 0x0040	/* Application code is DEC Kanji	*/

/*
 * ODL states
 */
#define AT_ODL_FIFO	0x0001
#define AT_ODL_LRU	0x0002
#define	AT_ODL_MAP	0x0004	/* Perform virtual to terminal code mapping */
#define AT_ODL_BUSY	0x0010
#define	AT_ODL_FLUSH	0x0020	/* Set if tty is flushing output queue	*/
#define	AT_ODL_VIRTUAL	0x0040	/* Virtual ODL mode flag		*/
#define	AT_ODL_MASK	0x007f

/*
 * KKC states
 */
#define	AT_KKC_DOING	 0x0001
#define	AT_KKC_MODIFYING 0x0002
#define	AT_KKC_BUSY	 0x0004
#define	AT_KKC_PENDING	 0x0008
#define	AT_KKC_LITERAL	 0x0010
#define	AT_KKC_XLITERAL	 0x0020
#define	AT_KKC_EXTENDED	 0x0100	/* In extended Japanese input mode */

/*
 * Character flag
 */
#define CHAR_NOECHO	 0x80000000

/*
 * Constants for ODL service structures
 */
#define ODL_MAXFNTS	256
#define	ODL_HASHSIZE	16   /* Size of ODL hash table - must be a power of 2 */

/*
 * SIM states
 */
#define	AT_SIM_ACTIVE	0x0001	/* SIM is activated	  */
#define	AT_SIM_STARTING	0x0002	/* SIM is being activated */
#define	AT_SIM_MASK	0x0003

#define	IS_SIM_ACTIVE(atp)	(IS_SIM_MODE(atp) && \
				((atp)->at_simstate & AT_SIM_ACTIVE))
#define	START_SIM(atp)		{(atp)->at_simstate |=  AT_SIM_ACTIVE ;}
#define	STOP_SIM(atp)		{(atp)->at_simstate &= ~AT_SIM_ACTIVE ;}

/*
 * at_simfuncs function index
 */
#define	SIM_PROCESS	0	/* Process input char in SIM mode */
#define	SIM_INIT	1	/* Start SIM mode		  */

/*
 * at_odlfuncs function index
 */
#define	ODL_PROCESS	0	/* Check output chars & perform ODL 	*/
#define	ODL_INIT	1	/* Initialize data structures for ODL	*/
#define	ODL_FLUSH	2	/* Flush the font lists 		*/

/*
 * macro functions (procedures)
 */
#define ATTY_BELL(tp, at'p')	(void)attyoutput(CTRL('g'), (tp), (atp))
#define	ATTY_UPRINTF(tp, atp, str) \
	{ \
	    attyoutstr(str, tp, atp)		; \
	    printf ("tty(%X) : %s", tp, str)	; \
	}

/*
 * EMPTY_QUEUE functions
 */
#define EMPTY_QUEUE(q_addr) \
	{ \
		while (getc(q_addr) >= 0) ; \
	}

#define EMPTY_WQUEUE(q_addr) \
	{ \
		while (getwc(q_addr) >= 0) ; \
	}

#define	EMPTY_BUFFER(b_addr) \
	{ \
		(b_addr)->c_cc = 0 ;		    \
		(b_addr)->c_ptr = (b_addr)->c_buf ; \
	}

#define	RUBOUT_WQUEUE(wclist, tp, atp)	\
	{ \
	    if  ( ((tp)->t_lflag & ECHO) && (atp)->at_clear_eol[0] && \
		(!((tp)->t_lflag & ECHOPRT)))	\
		atp->at_state |= AT_RUB_CLEAR;	\
	    while ((wclist).wc_cc) \
		attyrub(unputwc(&(wclist)), (tp), (atp)); \
	    if (atp->at_state & AT_RUB_CLEAR) {		  \
		ATTY_CLEAR_EOL(tp, atp);	\
		atp->at_state &= ~AT_RUB_CLEAR; \
	    } \
	}

/*
 * CATWQ_ECHOING
 */
#define CATWQ_ECHOING(fromq, toq, tp, atp) \
	{ \
	    register Wchar_t wc; \
	    while ((fromq).wc_cc) { \
		wc = getwc(&(fromq)); \
		if ((tp)->t_lflag & ECHO) \
		    (void)attyecho(wc, (tp), (atp)); \
		(void)putwc(wc, &(toq)); \
	    } \
	}

/*
 * Wide character macros
 */
#define WC_IS_2BSPACE(wc)	atty_is_2bspace(wc)
#define WC_IS_1BYTE(wc)		(((wc) & TTY_WCHARMASK) <= 0x00ff)
#define	WC_IS_SPACE(wc)		(((wc) == ' ') || ((wc) == '\t') || \
				 atty_is_2bspace(wc))

/*
 * The t_language field of the tty structure is used as a hook to
 * the atty structure.
 */
#define	t_atty		t_language
#define	kkseq_data_t	struct kkseq_data

/*
 * ATTY mode macros
 */
#define	IS_ASY_LINE(tp)		((tp)->t_line == ASYDISC)
#define	IS_ODL_MODE(atp)	((atp)->at_flags & AS_ODL)
#define	IS_KKC_MODE(atp)	((atp)->at_flags & AS_KKC)
#define	IS_JIN_MODE(atp)	((atp)->at_flags & AS_KKC)
#define	IS_SIM_MODE(atp)	((atp)->at_flags & AS_SIM)
#define	IS_DLC_MODE(atp)	((atp)->at_flags & AS_DLC)
#define	IS_CNV_MODE(atp)	((atp)->at_flags & AS_CONV_MASK	)
#define	IS_ACNV_MODE(atp)	((atp)->at_flags & AS_ACNV	)
#define	IS_TICNV_MODE(atp)	((atp)->at_flags & AS_TICNV	)
#define	IS_TOCNV_MODE(atp)	((atp)->at_flags & AS_TOCNV	)
#define	IS_VIRT_MODE(atp)	((atp)->at_odlstate & AT_ODL_VIRTUAL)

/*
 * ATTY magic number
 */
#define	ATTY_MAGIC	0x709394cd

/*===========================[ TYPES DEFINITIONS ]===========================*/


/*
 * The wide character type wchar_t has been declared unsigned short in types.h.
 * So it is re-declared here. It should be changed to
 *
 *	#define Wchar_t wchar_t
 *
 * if types.h is updated to define a 32-bit wide character.
 */
typedef	unsigned int	Wchar_t	;
typedef	int		Wint_t	;
typedef struct tty	tty_t	;
typedef struct clist	clist_t ;
typedef u_long		(*ulong_func)() ;

/*
 * A wclist structure is the head of a linked list queue
 * of wide characters.  The characters are stored in blocks
 * containing a link and WCBSIZE wide characters. 
 * The routines in atty_subr.c manipulate these structures.
 */
typedef struct wclist {
	int	 wc_cc;		/* Wide character count  */
	Wchar_t	*wc_cf;		/* Pointer to first char */
	Wchar_t	*wc_cl;		/* Pointer to last char  */
} wclist_t ;

/*
 * The wcblock is the actual data structure that contains the wide
 * characters. There is not need for wc_quote[] array as the wide
 * character encoding itself has extra space to record that bit.
 */
typedef struct wcblock {
	struct wcblock *wc_next;
	Wchar_t		wc_info[WCBSIZE];
} wcblock_t ;

typedef struct cblock	cblock_t ;

/*
 * A cbuffer structure is used to hold input character to be fed through the
 * mbtowc routine. The c_ptr pointer points to the beginning of valid data.
 */
typedef struct cbuffer {
	int	 c_cc		;	/* Character count   */
	u_char  *c_ptr		;	/* Character pointer */
	u_char	 c_buf[CBUFSIZ] ;	/* Character buffer  */
} cbuffer_t ;

/*
 * A cconv structure is used to hold code conversion specific information.
 * This structure will be used by both the BSD & STREAMS tty drivers.
 */
typedef struct cconv {
	u_char   *cc_in	     ;	/* Character input pointer  */
	u_char 	 *cc_out     ;	/* Character output pointer */
	cc_t	 *cc_cchars  ;	/* Control character list   */
	int	 *cc_tpstate ;	/* TP's state		    */
	tcflag_t *cc_tplflag ;	/* TP's lflag		    */
	tty_t	 *cc_tpaddr  ;	/* Address of tp	    */
	int	  cc_cc	     ;	/* Character count	    */
	int	  cc_state   ;	/* Code conversion state    */
	int	  cc_flags   ;	/* Code conversion flag	    */
	Wchar_t	  cc_lastwc  ;	/* Last wide character	    */
	int	  cc_si_len  ;	/* Length of shift_in []    */
	int	  cc_so_len  ;	/* Length of shift_out[]    */
	char	  cc_shift_in [AIO_ESCSEQSIZE] ;
	char	  cc_shift_out[AIO_ESCSEQSIZE] ;
} cconv_t ;

/*
 * Linked list of clauses (for Japan)
 */
typedef struct attyclause {
	struct attyclause *ac_next ;
	Wchar_t		  *ac_yomi ;
	Wchar_t		  *ac_kanji;
	unsigned int	   ac_ylen ;
	unsigned int	   ac_ytype;
	unsigned int	   ac_klen ;
} attyclause_t ;

/*
 * Linked list of dictionaries (for Japan)
 */
typedef struct attydict {
	struct attydict *ad_next;
	char		*ad_ptr ;
	unsigned int	 ad_type;
	unsigned int	 ad_size;
} attydict_t ;

/*
 * History structure
 */
typedef struct atty_history {
	short	h_last ;		/* Last entry number		*/
	short	h_now  ;
	short	h_loc  ;
	short	h_state;
	short	h_head ;		/* Head pointer to h_bufs	*/
	short	h_tail ;		/* Tail pointer to h_bufs	*/
	short	h_free ;		/* Amount of free buffer space	*/
	short	h_list[HISTMAX]	;
	Wchar_t h_keys[8]	;
	Wchar_t h_str [HISTLEN]	;
	Wchar_t h_bufs[HISTBUF]	;
} atty_history_t ;

/*
 * odl service structures
 */
typedef struct odl_fntnode {
	Wchar_t		    pcode   ;	/* Codeset value for DB searching */
	Wchar_t		    tcode   ;	/* Codeset value for DCS string	  */
	u_int		    codeset ;	/* Codeset number		  */
	u_int		    fontnum ;	/* Font number			  */
	struct odl_fntnode *in	    ;
	struct odl_fntnode *out	    ;
	struct odl_fntnode *prev    ;
	struct odl_fntnode *next    ;
} odl_fntnode_t ;

typedef struct odl_fnts {
	odl_fntnode_t  *out_fnts		;
	odl_fntnode_t  *in_fnts			;
	odl_fntnode_t  *free_fnts		;
	odl_fntnode_t   hash_fnts[ODL_HASHSIZE] ;
	int		total_fnts		;
} odl_fnts_t ;

/*
 * The atty structure
 */
typedef struct atty {
	int		 at_magic	;	/* Magic number for atty    */
	int		 at_flags	;	/* Asian terminal mode word */
	int		 at_state	;	/* State of atty	    */
	struct tty      *at_tty  	;	/* Pointer to tty structure */

	/* Code conversion routines */
	ulong_func	 at_timbtowc	;	/* Terminal input mb-to-wc  */
	ulong_func	 at_towctomb	;	/* Terminal output wc-to-mb */
	ulong_func	 at_aiwctomb	;	/* Appl. input wc-to-mb     */
	ulong_func	 at_aombtowc	;	/* Appl. output mb-to-mb    */
	ulong_func	 at_t_itoewc	;	/* Terminal int to ext wc   */
	ulong_func	 at_t_etoiwc	;	/* Terminal ext to int wc   */
	ulong_func	 at_a_itoewc	;	/* Appl. int to ext wc	    */
	ulong_func	 at_a_etoiwc	;	/* Appl. ext to int wc      */
	cconv_t		 at_cconv	;	/* Conversion state block   */

	/* New wide character queues and character buffers */
	wclist_t	 at_wrawq	;	/* Wide char raw queue	     */
	wclist_t	 at_wcanq	;	/* Wide char canonical queue */
	wclist_t	 at_woutq	;	/* Wide char output queue    */
	wclist_t	 at_wpendq	;	/* Wide char pending queue   */
	wclist_t	 at_wprehq	;	/* Wide char pre-henkan q.   */
	wclist_t	 at_wdispq	;	/* Wide char display queue   */
	wclist_t	 at_werawq	;	/* Extended raw queue	     */
	wclist_t	 at_wecanq	;	/* Extended canonical queue  */
	cbuffer_t	 at_ainpb	;	/* Appl. input buffer	     */
	cbuffer_t	 at_astib	;	/* Appl. STI input buffer    */
	cbuffer_t	 at_aoutb	;	/* Appl. output buffer       */
	cbuffer_t	 at_trawb	;	/* Terminal raw input buffer */
	cbuffer_t	 at_toutb	;	/* Terminal output buffer    */

	/* Fields for Kana-Kanji Conversion */
	int		 at_kkcstate	;	/* KKC state flag	    */
	short		 at_kkmapsize	;	/* Size of kkmap	    */
	short		 at_utxkkcerror	;	/* Utx-kkc error code	    */
	short		 at_clnow	;	/* Current clause number    */
	short		 at_clcnt	;	/* Current clause count	    */
	int		 at_jinkey	;	/* Extended inp. toggle key */
	kkseq_data_t	*at_kkmap	;	/* Pointer to kkmap	    */
	kkseq_data_t	*at_kkwp	;	/* Working pointer to kkmap */
	attydict_t	*at_dicts	;	/* List of dictionaries	    */
	attyclause_t	*at_clause	;	/* List of clauses	    */
	attyclause_t	*at_clause_old	;	/* Extended input	    */
	struct utx_tty	*at_utx_kkc	;	/* KKC			    */
	int_func	 at_endhenkan	;	/* End henkan function	    */
	int_func	*at_kkcfuncs	;	/* KKC			    */
	int_func	*at_extended	;	/* Extended input	    */
	Wchar_t		 at_kkckeys  [KKCCMDSIZE    ] ;
	char		 at_cmark_on [AIO_ESCSEQSIZE] ;
	char		 at_cmark_off[AIO_ESCSEQSIZE] ;
	char		 at_hmark_on [AIO_ESCSEQSIZE] ;
	char		 at_hmark_off[AIO_ESCSEQSIZE] ;

	/* Fields for On-Demand Loading */
	short		 at_odlstate	;	/* ODL state flag	    */
	short 		 at_utxodlerror	;	/* ODL			    */
	short		 at_odlmaxfnts	;	/* ODL			    */
	short		 at_odlfontnum	;	/* Current font number	    */
	short		 at_odlbusycnt  ;	/* utx request pending cnt  */
	short		 at_odlpathlen	;	/* ODL			    */
	char	        *at_odlpath	;	/* ODL			    */
	int_func        *at_odlfuncs	;	/* ODL			    */
	struct utx_tty  *at_utx_odl	;	/* ODL			    */
	odl_fnts_t 	*at_odlfnts	;	/* ODL			    */
	odl_fntnode_t	*at_odlnode	;	/* ODL			    */
	int		 at_odlchar	;	/* ODL			    */
	short		 at_aodlnum	;	/* No. of appl ODL areas    */
	short		 at_todlnum	;	/* No. of term ODL areas    */
	short		 at_podlnum	;	/* No. of phys ODL areas    */
	caddr_t		 at_aodlarea	;	/* Ptr to appl ODL areas    */
	caddr_t		 at_todlarea	;	/* Ptr to term ODL areas    */
	caddr_t		 at_podlarea	;	/* Ptr to phys ODL areas    */
	unsigned int	 at_physcode    ;	/* Current physical code    */
	unsigned int	 at_odlhit	;	/* For statistics	    */
	unsigned int	 at_odlmiss	;	/* For statistics	    */

	/* Field for Software Input Method */
	struct utx_tty  *at_utx_sim	;	/* sim			   */
	Wchar_t		*at_siminit	;	/* SIM init string	   */
	Wchar_t		*at_simexit	;	/* SIM exit string	   */
	Wchar_t		*at_simstart	;	/* SIM start string	   */
	Wchar_t		*at_simstop	;	/* SIM stop string	   */
	Wchar_t		*at_simctrlstr	;	/* SIM Control string	   */
	int_func        *at_simfuncs	;	/* SIM functions	   */
	char		*at_simpath     ;	/* Path of SIM directory   */
	char		 at_simclass[16];	/* Class name		   */
	short		 at_simctrllen  ;	/* Byte length of ctrl str */
	short		 at_simbusycnt  ;	/* Utx request pending cnt */
	short		 at_simpathlen  ;	/* Length of SIM directory */
	short		 at_simkey	;	/* Key to activate SIM	   */
	short		 at_simstate    ;	/* SIM state flag	   */
	short 		 at_utxsimerror	;	/* SIM has error	   */
	short		 at_simparam	;	/* SIM specific parameter  */

	/* Miscellaneous data */
	short		 at_termcode	;	/* Terminal (output) code  */
	short		 at_applcode	;	/* Application code	   */
	short		 at_termtype	;	/* Terminal type code	   */
	short		 at_langcode	;	/* Language code	   */

	/* History mechanism */
	atty_history_t	*at_hist	;	/* History structure	   */
	short		 at_histchar    ;	/* History activation char */
	char		 at_clear_eol[8];	/* Clear to EOL esc keys   */
} atty_t ;

/*
 * The cconv structure backward compatiblity macros
 */
#define	at_convstate	at_cconv.cc_state
#define	at_last_twc	at_cconv.cc_lastwc
#define	at_si_len	at_cconv.cc_si_len
#define	at_so_len	at_cconv.cc_so_len
#define	at_shift_in	at_cconv.cc_shift_in
#define	at_shift_out	at_cconv.cc_shift_out

/*==========================[ FUNCTION PROTOTYPES ]==========================*/

extern attywflush  (tty_t *tp, atty_t *atp) 				    ;
extern attywait    (tty_t *tp, atty_t *atp) 				    ;
extern attyflush   (tty_t *tp, long rw)					    ;
extern attyblock   (tty_t *tp)						    ;
extern attstart    (tty_t *tp) 						    ;
extern attystartcnv(tty_t *tp, atty_t *atp)				    ;
extern attioctl    (tty_t *tp, long com, caddr_t data, long flag)	    ;
extern attnread    (tty_t *tp, atty_t *atp)				    ;
extern attselect   (dev_t dev, short *events, short *revents, int scanning) ;
extern attyopen	   (dev_t dev, tty_t *tp, long flags)			    ;
extern attylclose  (tty_t *tp)						    ;
extern attyclose   (tty_t *tp) 						    ;
extern attymodem   (tty_t *tp, long flag) 				    ;
extern attypend    (tty_t *tp, atty_t *atp)				    ;
extern attyinput   (int  c, tty_t *tp)					    ;
extern attydoinput (Wchar_t wc, tty_t *tp, atty_t *atp)			    ;
extern attyoutput  (Wchar_t wc, tty_t *tp, atty_t *atp)			    ;
extern attread     (tty_t *tp, struct uio *uio, long flag)		    ;
extern attwrite	   (tty_t *tp, struct uio *uio, long flag)		    ;
extern attyrub     (Wchar_t wc, tty_t *tp, atty_t *atp)			    ;
extern attyrubo    (tty_t *tp, atty_t *atp, long cnt)			    ;
extern attyretype  (tty_t *tp, atty_t *atp)				    ;
extern attyecho    (Wchar_t wc, tty_t *tp, atty_t *atp)			    ;
extern attyoutstr  (char *cp, tty_t *tp, atty_t *atp)			    ;
extern attyoutwstr (Wchar_t *wcp, tty_t *tp, atty_t *atp)		    ;
extern attyinfo    (tty_t *tp, atty_t *atp)				    ;
extern attyoutint  (long n, long base, long min, tty_t *tp, atty_t *atp)    ;
extern atputchar   (Wchar_t wc, tty_t *tp, atty_t *atp)			    ;
extern getc	   (clist_t  *p )		 			    ;
extern getwc	   (wclist_t *wp)		 			    ;
extern wq_to_wb    (wclist_t *wq, Wchar_t *wcp, long wcc)		    ;
extern putc	   (long  c, clist_t  * p)				    ;
extern putwc	   (Wchar_t wc, wclist_t *wp)				    ;
extern wb_to_wq	   (Wchar_t *wcp, long wcc, wclist_t *wq)		    ;
extern b_to_wq	   (char *cp, long cc, wclist_t *wq)			    ;
extern catwq	   (wclist_t *from, wclist_t *to)			    ;
extern bgetc	   (cbuffer_t *b)					    ;
extern bputc	   (long c, cbuffer_t *b)				    ;
extern bunputc	   (cbuffer_t *b)					    ;
extern bputs	   (cbuffer_t *b, char *cp, long cc)			    ;
extern wstrncmp	   (Wchar_t *ws1, Wchar_t *ws2, long n)			    ;
extern Wchar_t *nextwc(wclist_t *wp, Wchar_t *wcp, Wchar_t *wc)		    ;

#ifdef	ASIAN_HISTORY
extern atty_history	 (Wchar_t wc, tty_t *tp, atty_t *atp) ;
extern atty_history_init (tty_t  *tp, atty_t *atp)	   ;
extern atty_history_break(tty_t  *tp, atty_t *atp)	   ;
extern atty_valid_at_hist(atty_history_t *ahistp)	   ;
extern ls_strsame	 (Wchar_t *ws1, char *s2)	   ;
extern ls_strpart	 (Wchar_t *ws1, char *s2)	   ;
#endif	/* ASIAN_HISTORY */

#endif	/* _SYS_ATTY_H_ */
