/*
 * *****************************************************************
 * *                                                               *
 * *    Copyright (c) Digital Equipment Corporation, 1991, 1996    *
 * *                                                               *
 * *   All Rights Reserved.  Unpublished rights  reserved  under   *
 * *   the copyright laws of the United States.                    *
 * *                                                               *
 * *   The software contained on this media  is  proprietary  to   *
 * *   and  embodies  the  confidential  technology  of  Digital   *
 * *   Equipment Corporation.  Possession, use,  duplication  or   *
 * *   dissemination of the software and media is authorized only  *
 * *   pursuant to a valid written license from Digital Equipment  *
 * *   Corporation.                                                *
 * *                                                               *
 * *   RESTRICTED RIGHTS LEGEND   Use, duplication, or disclosure  *
 * *   by the U.S. Government is subject to restrictions  as  set  *
 * *   forth in Subparagraph (c)(1)(ii)  of  DFARS  252.227-7013,  *
 * *   or  in  FAR 52.227-19, as applicable.                       *
 * *                                                               *
 * *****************************************************************
 */
/*
 * HISTORY
 */
/*
 */
/*	
 *	@(#)$RCSfile: ttty.h,v $ $Revision: 1.2.6.3 $ (DEC I18N) $Date: 1995/11/09 20:41:01 $
 */ 
/*
 * ttty.h
 *
 * Revision History:
 *
 * 17-Sep-92	W.M. Long
 *      First draft - this header file specifies all constants and data
 *	data structure used by the I18N single-byte BSD line discipline.
 */

#ifndef	_SYS_TTTY_H_
#define _SYS_TTTY_H_

#include <sys/tty.h>
#include <sys/uio.h>
#include <sys/aioctl.h>

/*====================[ LITERALS & MACROS DEFINITIONS ]======================*/

/*
 * Thai character classes
 */
typedef enum thai_class
{
    NON ,	/* Non-Thai character	*/
    CONS,	/* Consonant		*/
    LV  ,	/* Leading vowel	*/
    FV1 ,	/* Following vowel 1	*/
    FV2 ,	/* Following vowel 1	*/
    FV3 ,	/* Following vowel 1	*/
    CTRL,	/* Control character	*/
    BV1 ,	/* Below vowel 1	*/
    BV2 ,	/* Below vowel 2	*/
    BD  ,	/* Below diacritics	*/
    TONE,	/* Tone mark		*/
    AD1 ,	/* Above diacritics 1	*/
    AD2 ,	/* Above diacritics 2	*/
    AD3 ,	/* Above diacritics 3	*/
    AV1 ,	/* Above vowel 1	*/
    AV2 ,	/* Above vowel 2	*/
    AV3 	/* Above vowel 3	*/
} tclass_t ;

/*
 * Input sequence check result
 */
typedef enum isc_result
{
    _A	, 	/* Accept (next display cell)	*/
    _C	,	/* Accept (same display cell)	*/
    _S	,	/* Strict mode reject		*/
    _R	,	/* Reject			*/
    _X	,	/* Accept (non-display		*/
} isc_result_t ;

/*
 * Macros to process ISC results and tty mode flags
 */
#define	ISC_MODE(ttp)		((ttp)->tt_flags & TFLG_MODEMASK)
#define	ACCEPT1(m)		(((m) != _R))
#define	ACCEPT2(m)		(((m) != _S) && ((m) != _R))

/*
 * TSB tty state flags
 * The TS_ORDERING bit is set only when a consonant and a following
 * non-L3 character is read.
 */
#define	TTS_ISOPEN	0x01	/* TSB tty opened			 */
#define	TTS_ORDERING	0x02	/* In re-ordering mode			 */
#define	TTS_HISTORY	0x04
#define	TTS_LITERAL	0x08
#define	TTS_RUB_CLEAR	0x10

/*
 * Constants for history structure
 */
#define	HISTMAX		32	/* Number of history entries	*/
#define	HISTLEN		128	/* Max # of chars per entry	*/
#define	HISTBUF		(HISTMAX * HISTLEN / 4)
#define	HISTBEG		1	/* Beginning of history buffer	*/
#define	HISTEND		HISTBUF	/* End of history buffer	*/
#define	THS_LITERAL	0x01

/*
 * Aliases for queues and buffers
 */
#define RAWQ		(tp)->t_rawq
#define CANQ		(tp)->t_canq
#define OUTQ		(tp)->t_outq

/*
 * macro functions (procedures)
 */
#define TTY_BELL(t'p')	(void)ttyoutput(CTRL('g'), (tp))
#define	TSB_BELL(t'p')	(void)tsboutput(CTRL('g'), (tp))
#define	IS_TSB_LINE(tp)	((tp)->t_line == TSBDISC)

/*
 * EMPTY_QUEUE functions
 */
#define EMPTY_QUEUE(q_addr) \
	{ \
		while (getc(q_addr) >= 0) ; \
	}

/*
 * The t_language field of the tty structure is used as a hook to
 * the ttty structure.
 */
#define	t_ttty	t_language

/*
 * TTTY magic number
 */
#define	TTTY_MAGIC	0x70aabbcc

/*===========================[ TYPES DEFINITIONS ]===========================*/

typedef struct tty	tty_t	 ;
typedef struct clist	clist_t  ;
typedef struct cblock	cblock_t ;

/*
 * History structure
 */
typedef struct ttty_history {
	short	h_last ;		/* Last entry number		*/
	short	h_now  ;
	short	h_loc  ;
	short	h_state;
	u_char  h_keys[8]	;
	u_char  h_str [HISTLEN]	;
	u_char  h_list[HISTMAX][HISTLEN] ;
} ttty_history_t ;

/*
 * The TSB tty structure
 */
typedef struct ttty {
	int		 tt_magic	;	/* Magic number for ttty    */
	int		 tt_flags	;	/* TSB terminal mode word   */
	int		 tt_state	;	/* State of ttty	    */
	tty_t		*tt_tty  	;	/* Pointer to tty structure */
	cc_t		*tt_cc		;	/* Control character list   */

	/* Thai character class */
	int		 tt_leadclass	;	/* Class of leading char    */

	/* History mechanism */
	ttty_history_t	*tt_hist	;	/* History structure	   */
	short		 tt_histchar    ;	/* History activation char */
	char		 tt_clear_eol[8];	/* Clear to EOL esc keys   */

	/* Function pointers */
	int	(*tt_tsbecho  )(long, void *) ;	/* Echo routine		*/
	int	(*tt_tsboutput)(long, void *) ;	/* Output routine	*/
	int	(*tt_tsbinput )(long, void *) ;	/* Input routine	*/
	int	(*tt_tsbstart )(void *)	      ;	/* Output start routine	*/
} ttty_t ;

/*
 * Macros to use the function pointers
 */
#define	tsb_echo(x, y)	 (*ttp->tt_tsbecho  )(x, y)
#define	tsb_output(x, y) (*ttp->tt_tsboutput)(x, y)
#define	tsb_input(x, y)	 (*ttp->tt_tsbinput )(x, y)
#define	tsb_start(x)	 (*ttp->tt_tsbstart )(x)

/*=====================[ GLOBAL VARIABLE DECLARATIONS ]======================*/

extern const tclass_t	  tclass_table[256]   ;
extern const isc_result_t isc_table[17][17]   ;
extern const int	  is_reorderable[17]  ;
extern const int	  is_non_spacing[256] ;

/*==========================[ FUNCTION PROTOTYPES ]==========================*/

extern int	tttyflush   (tty_t *tp, long rw)			   ;
extern int	tttioctl    (tty_t *tp, long com, caddr_t data, long flag) ;
extern int	tttyopen    (dev_t dev, tty_t *tp, long flags)		   ;
extern int	tttyclose   (tty_t *tp) 				   ;
extern int	tttypend    (tty_t *tp, ttty_t *ttp)			   ;
extern void	tttysetclass(tty_t *tp, ttty_t *ttp)			   ;
extern int	tsb_isc_ok  (tty_t *tp, ttty_t *ttp, long c, int *swap)	   ;
extern int	tsb_history_init   (void *tp, ttty_t *ttp)		   ;
extern int	tsb_history (int c, void *tp, ttty_t *ttp)		   ;
extern int	tttyinput   (long c, tty_t *tp)				   ;
extern int	getc	    (clist_t  *p)		 		   ;
extern int	putc	    (long  c, clist_t *p)			   ;
extern char    *nextc	    (clist_t *p, char *cp, int *c)		   ;

#endif	/* _SYS_TTTY_H_ */
